/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#ifndef AM_H
#define AM_H

#include "mem.h"
#include "amdefs.h"

#define V_AN_I(_r) AREG(_r)

#define V_AN_PI_B_All(_r) (AREG(_r)++,\
					((_r)==7?(AREG(_r)++,AREG(_r)-2):(AREG(_r)-1)))
#define V_AN_PD_B_All(_r) (AREG(_r)--,\
					((_r)==7?(AREG(_r)--,AREG(_r)):AREG(_r)))

/* Assume r!=A7 and s!=1 */
#define V_AN_PI(_r,_s) (AREG(_r) += _s,AREG(_r)-(_s))
#define V_AN_PD(_r,_s) (AREG(_r) -= _s)

#define V_AN_D(_r) (pc+=2, AREG(_r)+LM_W(MEM(pc-2)))/* BUG:Alignment! */
#define V_ABS_W (pc+=2, LM_W(MEM(pc-2))) /* BUG: Alignment! */
#define V_ABS_L (pc+=4, LM_UL(MEM(pc-4))) /* BUG: Alignment! */

#define V_IMM_B (pc+=2, pc-1)
#define V_IMM_W (pc+=2, pc-2)
#define V_IMM_L (pc+=4, pc-4)

static INLINE L V_AN_R(int reg)
{
	int x = LM_W(MEM(pc));  /* BUG: Alignment! */
	int scale = BIT(10,x)<<1 | BIT(9,x); /* x1, x2, x4 or x8 */
	pc += 2;

	if (!BIT(8,x))
	{
/* brief extension word */
		if (!BIT(11,x))
			return AREG(reg)+SBITS(0,7,x)+(LDREG_W(UBITS(12,15,x)) << scale);
		return AREG(reg)+SBITS(0,7,x)+(DREG(UBITS(12,15,x)) << scale);
//			return AREG(reg)+SBITS(0,7,x)+LDREG_W(UBITS(12,15,x));
//		return AREG(reg)+SBITS(0,7,x)+DREG(UBITS(12,15,x));
	}
	else
	{
/* full extension word */
		long tmp_addr = 0;
		int mode,submode;
		/* BS: 0=base register added  1=suppressed */
		if (!BIT(7,x))
		{
			tmp_addr = AREG(reg);
		}
		mode = UBITS(4,5,x);
		switch(mode)
		{
		case 0:
			/*reserved*/
			pc -= 2;
			ILLEGAL();
			break;
		case 1:
			/* no base displacement */
			break;
		case 2:
			/* word base displacement */
			tmp_addr += (long)LM_W(MEM(pc));
			pc += 2;
			break;
		case 3:
			/* long base displacement */
			tmp_addr += (long)LM_L(MEM(pc));
			pc += 4;
			break;
		}
		mode = BIT(2,x) | BIT(6,x)<<1;
		submode = UBITS(0,1,x);
		/* bit6=IS: 0=eval&add index operand  1=suppress */
		switch(mode)
		{
		case 0:
			if (!BIT(11,x))
				tmp_addr += (LDREG_W(UBITS(12,15,x)) << scale);
			else
				tmp_addr += (DREG(UBITS(12,15,x)) << scale);
			switch(submode)
			{
			case 0:
				/* with index and no memory indirect */
				break;
			case 1:
				/* preindexed with null outer displ */
				tmp_addr = LM_L(MEM(tmp_addr));
				break;
			case 2:
				/* preindexed with word outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* preindexed with long outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 1:
			tmp_addr = LM_L(MEM(tmp_addr));
			if (!BIT(11,x))
				tmp_addr += (LDREG_W(UBITS(12,15,x)) << scale);
			else
				tmp_addr += (DREG(UBITS(12,15,x)) << scale);
			switch(submode)
			{
			case 0:
				/* reserved */
				pc -= 2;
				ILLEGAL();
				break;
			case 1:
				/* postindexed with null outer displ */
				break;
			case 2:
				/* postindexed with word outer displ */
				tmp_addr += (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* postindexed with long outer displ */
				tmp_addr += (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 2:
			switch(submode)
			{
			case 0:
				/* no index, no memory indirect */
				break;
			case 1:
				/* no index, memory indirect with null outer displ */
				tmp_addr = LM_L(MEM(tmp_addr));
				break;
			case 2:
				/* no index, memory indirect with word outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* no index, memory indirect with long outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 3:
			/* reserved */
			pc -= 2;
			ILLEGAL();
			break;
		}
		return tmp_addr;
	}
}

#define V_PC_D (pc += 2, pc-2+LM_W(MEM((pc-2))))/* BUG: Alignment! */
#define V_PC_R xV_PC_R()

static INLINE L xV_PC_R(void)
{
	int x = LM_W(MEM(pc));  /* BUG: Alignment! */
	int scale = BIT(10,x)<<1 | BIT(9,x); /* x1, x2, x4 or x8 */
	pc += 2;

	if (!BIT(8,x))
	{
/* brief extension word */
		if (!BIT(11,x))
			return pc-2+SBITS(0,7,x)+(LDREG_W(UBITS(12,15,x)) << scale);
		return pc-2+SBITS(0,7,x)+(DREG(UBITS(12,15,x)) << scale);
//			return AREG(reg)+SBITS(0,7,x)+LDREG_W(UBITS(12,15,x));
//		return AREG(reg)+SBITS(0,7,x)+DREG(UBITS(12,15,x));
	}
	else
	{
/* full extension word */
		long tmp_addr = 0;
		int mode,submode;
		/* BS: 0=base register added  1=suppressed */
		if (!BIT(7,x))
		{
			tmp_addr = pc-2;
		}
		mode = UBITS(4,5,x);
		switch(mode)
		{
		case 0:
			/*reserved*/
			pc -= 2;
			ILLEGAL();
			break;
		case 1:
			/* no base displacement */
			break;
		case 2:
			/* word base displacement */
			tmp_addr += (long)LM_W(MEM(pc));
			pc += 2;
			break;
		case 3:
			/* long base displacement */
			tmp_addr += (long)LM_L(MEM(pc));
			pc += 4;
			break;
		}
		mode = BIT(2,x) | BIT(6,x)<<1;
		submode = UBITS(0,1,x);
		/* bit6=IS: 0=eval&add index operand  1=suppress */
		switch(mode)
		{
		case 0:
			if (!BIT(11,x))
				tmp_addr += (LDREG_W(UBITS(12,15,x)) << scale);
			else
				tmp_addr += (DREG(UBITS(12,15,x)) << scale);
			switch(submode)
			{
			case 0:
				/* with index and no memory indirect */
				break;
			case 1:
				/* preindexed with null outer displ */
				tmp_addr = LM_L(MEM(tmp_addr));
				break;
			case 2:
				/* preindexed with word outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* preindexed with long outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 1:
			tmp_addr = LM_L(MEM(tmp_addr));
			if (!BIT(11,x))
				tmp_addr += (LDREG_W(UBITS(12,15,x)) << scale);
			else
				tmp_addr += (DREG(UBITS(12,15,x)) << scale);
			switch(submode)
			{
			case 0:
				/* reserved */
				pc -= 2;
				ILLEGAL();
				break;
			case 1:
				/* postindexed with null outer displ */
				break;
			case 2:
				/* postindexed with word outer displ */
				tmp_addr += (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* postindexed with long outer displ */
				tmp_addr += (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 2:
			switch(submode)
			{
			case 0:
				/* no index, no memory indirect */
				break;
			case 1:
				/* no index, memory indirect with null outer displ */
				tmp_addr = LM_L(MEM(tmp_addr));
				break;
			case 2:
				/* no index, memory indirect with word outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_W(MEM(pc));
				pc += 2;
				break;
			case 3:
				/* no index, memory indirect with long outer displ */
				tmp_addr = LM_L(MEM(tmp_addr)) + (long)LM_L(MEM(pc));
				pc += 4;
				break;
			}
			break;
		case 3:
			/* reserved */
			pc -= 2;
			ILLEGAL();
			break;
		}
		return tmp_addr;
	}
}

#define AN_I(_r) MEM(V_AN_I(_r))
#define AN_PI(_r,_s) MEM(V_AN_PI(_r,_s))
#define AN_PD(_r,_s) MEM(V_AN_PD(_r,_s))
#define AN_D(_r) MEM(V_AN_D(_r))
#define ABS_W MEM(V_ABS_W)
#define ABS_L MEM(V_ABS_L)
#define AN_R(_r) MEM(V_AN_R(_r))
#define PC_D MEM(V_PC_D)
#define PC_R MEM(V_PC_R)

#endif /* AM_H */






