/*
 * Network access routines for emulated Atari
 * ( BETA CODE )
 *
 * Interface based on MPCStiK
 * Same TSR client used on Atari side than MPCStiK
 * This code is beta, and it probably should get some infos
 * from configure... This has been (half-)tested on Ubuntu x86.
 * (functions with comments have been tested)
 */

#include "defs.h"
#include "cpu.h"
#include "gendefs.h"
#include "mem.h"
#include "gemdos.h"
#include "native.h"
#include "net.h"
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

/* uncomment NETLOG to display debug infos */
//#define NETLOG

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define P_ICMP	1
#define P_TCP	6
#define P_UDP	17

/* TCP connection states    */
#define TCLOSED     0   /* No connection.  Null, void, absent, .....        */
#define TLISTEN     1   /* Wait for remote request.  Not used here.         */
#define TSYN_SENT   2   /* Connect request sent.  Await matching request    */
#define TSYN_RECV   3   /* Wait for connection ACK.  (Listener only??)      */
#define TESTABLISH  4   /* Connection is established.  Handshake completed  */
#define TFIN_WAIT1  5   /* Await termination request or acknowledgement     */
#define TFIN_WAIT2  6   /* Await termination request.                       */
#define TCLOSE_WAIT 7   /* Await termination request from local user        */
#define TCLOSING    8   /* Await termination acknowledge from remote TCP    */
#define TLAST_ACK   9   /* Await acknowledgement of terminate request sent  */
#define TTIME_WAIT 10   /* Delay to ensure remote has received term' ACK    */

/* Addressing information blobk */
typedef struct cab
{
	unsigned short	lport;
	unsigned short	rport;
	unsigned long	rhost;
	unsigned long	lhost;
} CAB;

/* Connection information block */
typedef struct cib
{
	unsigned short	protocol;
	//	CAB		adress;
	unsigned int	lport;
	unsigned int	rport;
	unsigned long	rhost;
	unsigned long	lhost;
	unsigned int	status;
} CIB;
typedef struct
{
	short open;	// is_opened flag
	int s;	// Socket Descriptor
	CIB info;
} fh_element;

#define FH_ARRAY_MAX 128
fh_element fh_array[FH_ARRAY_MAX];

int get_free_handle(void)
{
	int ret = -1;
	int i;
	for(i=1; i<FH_ARRAY_MAX; i++)
	{
		if (fh_array[i].open == FALSE)
		{
			ret = i;
			break;
		}
	}
	return ret;
}

void fill_CIB (int handle)
{
	int s;
	int len = sizeof(struct sockaddr_in);
	struct sockaddr_in laddr,raddr;
	int llen,rlen;
	unsigned long ip;

	s = fh_array[handle].s;
	memset(&laddr, 0, len);
	memset(&raddr, 0, len);
	llen = len;
	rlen = len;

	/* get local part */
	if (getsockname(s, (struct sockaddr*)&laddr, &llen) == 0)
	{
		memcpy( &ip, &laddr.sin_addr, 4);
		fh_array[handle].info.lhost = ip;
		fh_array[handle].info.lport = laddr.sin_port; 
	}
	else
	{
		fh_array[handle].info.lhost = 0;
		fh_array[handle].info.lport = 0;
	}
	/* get remote part */
	if (getsockname(s, (struct sockaddr*)&raddr, &rlen) == 0)
	{
		memcpy( &ip, &raddr.sin_addr, 4);
		fh_array[handle].info.rhost = ip;
		fh_array[handle].info.rport = raddr.sin_port; 
	}
	else
	{
		fh_array[handle].info.rhost = 0;
		fh_array[handle].info.rport = 0;
	}

	return;
}

/* structs shared with atari side */
typedef struct msock_info
{
	char *name;     /* Hier wird der Hostname abgelegt */
	char *ip;       /* Hier wird die IP abgelegt */
	short ver;      /* Hier landet die WINSOCK Versionsnummer */
	short res1;     /* Reserviert */
	char *work;     /* Reserviert */
	short res3;     /* Reserviert */
} MSOCK_INFO;


/* private functions */

static void stonxnet_init(void)
{
	MSOCK_INFO *stadr = MEM(LDREG_LONG(8)); /* get struct address from a0 */
	char *buf1 = MEM(LM_UL(&stadr->name));
	char *buf2 = MEM(LM_UL(&stadr->ip));
	struct hostent *host;
	char *szIPAddress;
	int idx;
#ifdef NETLOG
	printf("atari requested initializing of stonxnet...\n");
#endif
	gethostname(buf1,256);
	host = gethostbyname(buf1);
	if (host != NULL)
	{
		szIPAddress = inet_ntoa (*(struct in_addr *)*host->h_addr_list);
		sprintf(buf2 , "%s%c%c" , szIPAddress,0,0);
	}
	else
	{
		sprintf(buf2 , "127.0.0.1%c%c",0,0);
	}
	for(idx=0; idx<FH_ARRAY_MAX; idx++)
	{
		fh_array[idx].open = FALSE;
	}
	SDREG_LONG(0,0); /* return 0 in d0 */
	return;
}

static void stonxnet_exit(void)
{
	int idx;
#ifdef NETLOG
	printf("atari requested deinitializing of stonxnet...\n");
#endif
	for(idx=0; idx<FH_ARRAY_MAX; idx++)
	{
		if (fh_array[idx].open != FALSE)
		{
			close(fh_array[idx].s);
			fh_array[idx].open = FALSE;
		}
	}
	SDREG_LONG(0,0); /* return 0 in d0 */
	return;
}

static void stonxnet_log(void)
{
	char *stmsg = MEM(LDREG_LONG(8)); /* get msg address from a0 */
	printf("stonxnet: %s\n",stmsg);
	SDREG_LONG(0,0); /* return 0 in d0 */
	return;
}

static void stonxnet_socket(void)
{
	int s;
	int handle;

	handle = get_free_handle();
#ifdef NETLOG
	printf("stonxnet new socket, free handle=%d : ", handle);
#endif
	if (handle > -1)
	{
		s = socket(LDREG_LONG(0),LDREG_LONG(1),LDREG_LONG(2)); //(d0,d1,d2)
		if (s == -1)
		{
#ifdef NETLOG
			printf("failed create socket\n");
#endif
			SDREG_LONG(0,-1);
		}
		else
		{
			fh_array[handle].open = TRUE;
			fh_array[handle].s = s;
			switch(LDREG_LONG(1))
			{
				case SOCK_STREAM:
					fh_array[handle].info.protocol = P_TCP;
					break;
				case SOCK_DGRAM:
					fh_array[handle].info.protocol = P_UDP;
					break;
				default:
					fh_array[handle].info.protocol = P_UDP;
					break;
			}
#ifdef NETLOG
			printf("created socket %d\n", s);
#endif
			SDREG_LONG(0,handle);
		}
	}
	else
	{
#ifdef NETLOG
		printf("no free slot\n");
#endif
		SDREG_LONG(0,-35); // d0 = ENHNDL
	}
	return;
}

static void stonxnet_close(void)
{
	int handle = LDREG_LONG(0);
	int s;
#ifdef NETLOG
	printf("stonxnet closing handle %d : ", handle);
#endif
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		if (close(s) == -1)
		{
#ifdef NETLOG
			printf(" error closing socket %d\n",s);
#endif
			SDREG_LONG(0,-14);	// d0 = E_BADCLOSE
		}
		else
		{
#ifdef NETLOG
			printf(" closed socket %d\n",s);
#endif
			fh_array[handle].info.status = TCLOSED;
			fh_array[handle].open = FALSE;
			SDREG_LONG(0,0);
		}
	}
	else
	{
#ifdef NETLOG
		printf(" error, bad handle\n");
#endif
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_connect(void)
{
	int handle = LDREG_LONG(0);
	int s;
	short port;
	struct sockaddr_in *saddr, addr;
	memset(&addr, 0, sizeof(struct sockaddr_in));

#ifdef NETLOG
	printf("stonxnet connect handle %d : ",handle);
#endif
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		saddr = (struct sockaddr_in*) MEM(LDREG_LONG(8)); /* get struct addr from a0 */

		addr.sin_family = LM_UW(&saddr->sin_family);
		/* fill local sockaddr_in from atari's one (which is in network order)*/
		addr.sin_port = saddr->sin_port;
		addr.sin_addr = saddr->sin_addr;

		port = ntohs(addr.sin_port);
#ifdef NETLOG
		printf(" to %08x:%d : ",LM_UL(&saddr->sin_addr),port);
#endif
		if (connect(s, (struct sockaddr*)&addr, sizeof(struct sockaddr)) == -1)
		{
			if (errno == EACCES)
			{
#ifdef NETLOG
				printf(" refused\n");
#endif
				fh_array[handle].info.status = TCLOSED;
				SDREG_LONG(0,-7); // d0 = E_REFUSE
			}
			else if (errno == EINPROGRESS)
			{
#ifdef NETLOG
				printf(" pending\n");
#endif
				fh_array[handle].info.status = TESTABLISH;
				fill_CIB(handle);
				SDREG_LONG(0,0);
			}
			else
			{
#ifdef NETLOG
				printf(" failed\n");
#endif
				fh_array[handle].info.status = TCLOSED;
				SDREG_LONG(0,-1);
			}
		}
		else
		{
#ifdef NETLOG
			printf(" ok\n");
#endif
			fh_array[handle].info.status = TESTABLISH;
			fill_CIB(handle);
			SDREG_LONG(0,0);
		}
	}
	else
	{
#ifdef NETLOG
		printf(" bad handle\n");
#endif
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
}

static void stonxnet_listen(void)
{
	int handle = LDREG_LONG(0);
	int s;
	struct sockaddr_in *saddr,addr;
	memset(&addr, 0, sizeof(struct sockaddr_in));
	int optVal = 1;
	int optLen = sizeof(optVal);

	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		saddr = (struct sockaddr_in*) MEM(LDREG_LONG(8)); /* get struct addr from a0 */

		addr.sin_family = LM_UW(&saddr->sin_family);
		/* fill local sockaddr_in from atari's one (which is in network order)*/
		addr.sin_port = saddr->sin_port;
		/*bind any address rather than addr.sin_addr = saddr->sin_addr; */
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		/* ask for socket to reuse address */
		setsockopt(s, SOL_SOCKET, SO_REUSEADDR,  (char*)optVal, optLen);
		if (bind(s, (struct sockaddr*)&addr, sizeof(struct sockaddr)) != 0)
		{
			//close(s);
			fh_array[handle].open = FALSE;
			SDREG_LONG(0,-1);
			return;
		}
		if (listen(s,5) != 0)
		{
			fh_array[handle].info.status = TCLOSED;
			SDREG_LONG(0,-1);
		}
		else
		{
			fill_CIB(handle);
			fh_array[handle].info.status = TLISTEN;
			SDREG_LONG(0,0);
		}
	}
	else
	{
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_accept(void)
{
	int handle = LDREG_LONG(0);
	int s;
	int acceptsocket;

	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		if(fh_array[handle].info.status == TLISTEN)
		{
			acceptsocket = accept(s,NULL,0);
			if (acceptsocket == -1)
			{
				if (errno==EAGAIN) // || errno==EWOULDBLOCK)
				{
					/* no connection yet */
					SDREG_LONG(0,TLISTEN);
				}
				else
				{
					/* error */
					SDREG_LONG(0,-1);
				}
			}
			else
			{
				/* success connecting, we got a new socket */
				fh_array[handle].info.status = TESTABLISH;
				fh_array[handle].s = acceptsocket;
				close(s); /* original socket still exists, close it */
				fill_CIB(handle);
				SDREG_LONG(0,TESTABLISH);
			}
		}
		else
		{
			SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
		}
	}
	else
	{
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_send(void)
{
	int handle = LDREG_LONG(0);
	int s;
	int acceptsocket;
	char *buf;
	int len;
	int res;

#ifdef NETLOG
	printf("stonxnet send handle %d : ",handle);
#endif
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		len = LDREG_LONG(1);
		buf = MEM(LDREG_LONG(8));
#ifdef NETLOG
		printf("sending %d bytes from st $%08x : ", len, LDREG_LONG(8));
#endif
		res = send(s,buf,len,0);
		if (res == -1)
		{
			if (errno == EAGAIN)
			{
#ifdef NETLOG
				printf("queue full, wait\n");
#endif
				SDREG_LONG(0,-1); //E_OBUFFULL
			}
			else if (errno == ECONNRESET)
			{
#ifdef NETLOG
				printf("cnx lost\n");
#endif
				SDREG_LONG(0,-4); //E_RRESET
			}
			else
			{
#ifdef NETLOG
				printf("error %d\n", errno);
#endif
				SDREG_LONG(0,-1);
			}
		}
		else
		{
#ifdef NETLOG
			printf("sent %d\n",res);
#endif
			SDREG_LONG(0,res);
		}
	}
	else
	{
#ifdef NETLOG
		printf("bad handle\n");
#endif
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_recv(void)
{
	int handle = LDREG_LONG(0);
	int s;
	int acceptsocket;
	char *buf;
	int len;
	int res;

#ifdef NETLOG
	printf("stonxnet recv handle %d : ",handle);
#endif
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		len = LDREG_LONG(1);
		buf = MEM(LDREG_LONG(8));
		res = recv(s,buf,len,0);
		if (res == -1)
		{
			if (errno == EAGAIN)
			{
#ifdef NETLOG
				printf("queue empty, wait\n");
#endif
				SDREG_LONG(0,-1); //E_OBUFFULL
			}
			else if (errno == ECONNRESET)
			{
#ifdef NETLOG
				printf("cnx lost\n");
#endif
				SDREG_LONG(0,-4); //E_RRESET
			}
			else
			{
#ifdef NETLOG
				printf("error %d\n", errno);
#endif
				SDREG_LONG(0,-1);
			}
		}
		else
		{
#ifdef NETLOG
			printf("got %d\n",res);
#endif
			SDREG_LONG(0,res);
		}
	}
	else
	{
#ifdef NETLOG
		printf("bad handle\n");
#endif
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_select(void)
{
	SDREG_LONG(0,-1);
	return;
}

static void stonxnet_inet_addr(void)
{
	char *ip = MEM(LDREG_LONG(8)); /* get string from a0 */
	SDREG_LONG(0,inet_addr(ip));
#ifdef NETLOG
	printf("stonxnet inetaddr '%s' = %08lx\n", ip, LDREG_LONG(0));
#endif
	return;
}

static void stonxnet_resolve(void)
{
	char *name = MEM(LDREG_LONG(8)); /* get name from a0 */
	struct hostent *host;
	unsigned long ip;

	if (name == NULL)
	{
#ifdef NETLOG
		printf("stonxnet resolve NULL\n");
#endif
		SDREG_LONG(0,-17); /* E_CANTRESOLVE */
	}
	else
	{
#ifdef NETLOG
		printf("stonxnet resolve '%s' : ",name);
#endif
		host = gethostbyname(name);
		if (host == NULL)
		{
#ifdef NETLOG
			printf("failed\n");
#endif
			SDREG_LONG(0,-17); /* E_CANTRESOLVE */
		}
		else
		{
			memcpy(&ip, host->h_addr_list[0], 4);
#ifdef NETLOG
			printf("found %08x\n", LM_UL(&ip));
#endif
			SDREG_LONG(0,LM_UL(&ip));
		}
	}
}

static void stonxnet_readable(void)
{
	int handle = LDREG_LONG(0);
	int s;
	fd_set rdfs;
	struct timeval tv;
	int retval;

	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		FD_ZERO(&rdfs);
		FD_SET(s, &rdfs);
		tv.tv_sec = 0;
		tv.tv_usec = 25;
		retval = select(s+1, &rdfs, NULL, NULL, &tv);
		if (retval == -1)
		{
			SDREG_LONG(0,-1);
		}
		else
		{
			/* there was only 1 fd in the set, so retval is either 0 or 1 */
			SDREG_LONG(0,retval);
		}
	}
	else
	{
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_writeable(void)
{
	int handle = LDREG_LONG(0);
	int s;
	fd_set wrfs;
	struct timeval tv;
	int retval;

	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		FD_ZERO(&wrfs);
		FD_SET(s, &wrfs);
		tv.tv_sec = 0;
		tv.tv_usec = 10;
		retval = select(s+1, &wrfs, NULL, NULL, &tv);
		if (retval == -1)
		{
			SDREG_LONG(0,-1);
		}
		else
		{
			/* there was only 1 fd in the set, so retval is either 0 or 1 */
			SDREG_LONG(0,retval);
		}
	}
	else
	{
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_bytecount(void)
{
	int handle = LDREG_LONG(0);
	int s;
	fd_set rdfs;
	struct timeval tv;
	int retval;
	unsigned long count;

#ifdef NETLOG
	printf("stonxnet bytecount handle %d : ",handle);
#endif
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		FD_ZERO(&rdfs);
		FD_SET(s, &rdfs);
		tv.tv_sec = 0;
		tv.tv_usec = 10;
		retval = select(s+1, &rdfs, NULL, NULL, &tv);
		if (retval == -1)
		{
#ifdef NETLOG
			printf("select error %d\n", errno);
#endif
			SDREG_LONG(0,-1);
		}
		else if (retval > 0)
		{
			retval = ioctl(s, FIONREAD, &count);
			if (retval == -1)
			{
#ifdef NETLOG
				printf("select 0, ioctl 0\n");
#endif
				SDREG_LONG(0,-1);
			}
			else
			{
				if (count == 0)
				{
#ifdef NETLOG
					printf("count 0, check status : ");
#endif
					char szBuffer[1] = "";
					retval = recv(s, szBuffer, 1, MSG_PEEK);
					if (retval == -1)
					{
						if ((errno == ECONNRESET) || (errno == ENOTCONN))
						{
#ifdef NETLOG
							printf("cnx lost\n");
#endif
							//close(s);
							fh_array[handle].info.status= TCLOSED;
							SDREG_LONG(0,-3); /* E_EOF */
						}
						else
						{
#ifdef NETLOG
							printf("error %d\n", errno);
#endif
							SDREG_LONG(0,0);
						}
					}
					else
					{
#ifdef NETLOG
							printf("OK\n");
#endif
							SDREG_LONG(0,0);
					}
				}
				else
				{
					if (count > 32767)
					{
						count = 32767;
					}
#ifdef NETLOG
					printf("count = %d\n", count);
#endif
					SDREG_LONG(0,count);
				}
			}
		}
		else
		{
#ifdef NETLOG
			printf("select says no data\n");
#endif
			SDREG_LONG(0,0);
		}
	}
	else
	{
#ifdef NETLOG
		printf("bad handle\n");
#endif
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_socketinfo(void)
{
	int handle = LDREG_LONG(0);
	CIB *info;

	if (fh_array[handle].open == TRUE)
	{
		info = (CIB*) MEM(LDREG_LONG(8));
		if (info != NULL)
		{
			SM_UW(MEM(&info->protocol), fh_array[handle].info.protocol);
			STORE_UL(MEM(&info->lhost), fh_array[handle].info.lhost);
			STORE_UW(MEM(&info->lport), fh_array[handle].info.lport);
			STORE_UL(MEM(&info->rhost), fh_array[handle].info.rhost);
			STORE_UW(MEM(&info->rport), fh_array[handle].info.rport);
			SM_UW(MEM(&info->status), fh_array[handle].info.status);
			SDREG_LONG(0,fh_array[handle].info.status);
		}
		else
		{
			SDREG_LONG(0,-1);
		}
	}
	else
	{
		SDREG_LONG(0,-9);	// d0 = E_BADHANDLE
	}
	return;
}

static void stonxnet_gethost(void)
{
	int handle = LDREG_LONG(0);

	if (fh_array[handle].open == TRUE)
	{
		/* info values stored as network order = bigendian = atari) */
		/* so no conversion ? */
		if (LDREG_LONG(1) == 1)
		{
			SDREG_LONG(0, LM_UL(&fh_array[handle].info.rhost));
		}
		else
		{
			SDREG_LONG(0,LM_UL(&fh_array[handle].info.lhost));
		}
	}
	else
	{
		SDREG_LONG(0,0);
	}
	return;
}

static void stonxnet_getport(void)
{
	int handle = LDREG_LONG(0);
	int mode = LDREG_LONG(1);

	if (fh_array[handle].open == TRUE)
	{
		/* info values stored as network order = bigendian = atari) */
		/* so no conversion ? */
		if (mode == 0)
		{
			SDREG_LONG(0, LM_UW(&fh_array[handle].info.lport));
		}
		else if (mode == 1)
		{
			SDREG_LONG(0, LM_UW(&fh_array[handle].info.rport));
		}
		else if (mode == 2)
		{
			SDREG_LONG(0,fh_array[handle].info.status);
		}
		else
		{
			SDREG_LONG(0,0);
		}
	}
	else
	{
		SDREG_LONG(0,0);
	}
	return;
}

static void stonxnet_ioctl(void)
{
	int handle = LDREG_LONG(0);
	int s;
	int ret;
	unsigned long ul = 1;

	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		ret = ioctl(s, FIONBIO, &ul);
	}
	SDREG_LONG(0,0);
	return;
}

static void stonxnet_setopt(void)
{
	int handle = LDREG_LONG(0);
	int s;
	int ret;
	int optname,optlen;
	char *optval = MEM(LDREG_LONG(8));
	optname = LDREG_LONG(1);
	optlen = LDREG_LONG(2);
	if (fh_array[handle].open == TRUE)
	{
		s = fh_array[handle].s;
		ret = setsockopt(s, IPPROTO_IP, optname, optval, optlen);
	}
	SDREG_LONG(0,0);
	return;
}

/* public function */
void call_net(int func)
{
	switch(func)
	{
		case 0:
			stonxnet_init();
			break;
		case 1:
			stonxnet_exit();
			break;
		case 2:
			stonxnet_log();
			break;
		case 3:
			stonxnet_socket();
			break;
		case 4:
			stonxnet_close();
			break;
		case 5:
			stonxnet_connect();
			break;
		case 6:
			stonxnet_listen();
			break;
		case 7:
			stonxnet_accept();
			break;
		case 8:
			stonxnet_send();
			break;
		case 9:
			stonxnet_recv();
			break;
		case 10:
			stonxnet_select();
			break;
		case 11:
			stonxnet_inet_addr();
			break;
		case 12:
			stonxnet_resolve();
			break;
		case 13:
			stonxnet_readable();
			break;
		case 14:
			stonxnet_writeable();
			break;
		case 15:
			stonxnet_bytecount();
			break;
		case 16:
			stonxnet_socketinfo();
			break;
		case 17:
			stonxnet_gethost();
			break;
		case 18:
			stonxnet_getport();
			break;
		case 19:
			stonxnet_ioctl();
			break;
		case 20:
			stonxnet_setopt();
			break;
		default:
			pc -= 6;
			ILLEGAL();
	}
	return;
}

